# About Code
The proposed algorithm for cyclic two-body approximation is given in `partial_two_body_app.jl`, which is implemented in Julia 1.8.
On the command line, we can use the algorithm as follows.
```
cd src
$ julia
julia> include("partial_two_body_app.jl")
julia> X = rand(10,10,10);
julia> Xr = b2_decomp(X,verbose=true)
```

The output `Xr` approximates `X`. Note that our model does not require target rank since our model is rank-free.


Our experiments on synthetic datasets can be performed from the command line as follows.
```
julia> include("exp_synt.jl")
julia> experiment_synthetic2(5,j=30,ring_input=true,true_ring_rank=15)
julia> experiment_synthetic2(6,j=20,ring_input=true,true_ring_rank=10)
julia> experiment_synthetic2(5,j=30,ring_input=false)
julia> experiment_synthetic2(6,j=20,ring_input=false)
```

Results for synthetic datasets obtained by the above commands correspond to Fig. 4(a)(b)(c)(d) in our paper.
Results will be saved in `../result` as jld2 files.
​
The following commands
```
julia> include("plot_synt.jl")
julia> syn2(ring_input=true, D=5,j=30,true_ring_rank=15, accum=true)
julia> syn2(ring_input=true, D=6,j=20,true_ring_rank=10, accum=true)
julia> syn2(ring_input=false,D=6,j=20,accum=true)
julia> syn2(ring_input=false,D=5,j=30,accum=true)
```

make pdf images from jld2 files. The generated pdf files will be saved in `../figs/syn2`.
We can modify plot conditions by editing the file `plot_config.jl`.
​

For experiments with real datasets, real-world datasets have to be stored in `../../../datasets/tensors/` as jld2 files in advance. To access the files, `loader_data.jl` can be used.
​Please refer to Appendix for the information on how to obtain real datasets.
Our experiments on real datasets can be performed from the command line as follows.
```
julia exp_real.jl
```

Results for real datasets obtained by the above commands correspond to Fig. 5 in our paper.
Results will be saved in `../result` as jld2 files.
The following command
```
julia plot_real.jl
```

makes pdf images from jld2 files. The generated pdf files will be saved in `../figs/real1`.